//	Copyright  2008.	GSI Technology
//						Jeff Daugherty
//						apps@gsitechnology.com
// 1.0   01/29/08       1) Created model and removed UB/LB control


//`define     SP8NS
//`define     SP9NS
`define     SP10NS
//`define     SP12NS
//`define     SP15NS


`timescale 1ns / 100ps

module GS78132A (A,             // address
		 DQ,             // data
		 nCE,            // chip enable
		 nWE,            // write enable
		 nOE             // output enable
		 );

   parameter	mem_size 	= 8192*1024/32,
    		  A_size 	= 18,
		  DQ_size 	= 32;

   input [A_size-1:0] A; 			// address
   inout [DQ_size:1]  DQ;		    // data
   input 	      nCE;		// chip enable
   input 	      nWE;		// write enable
   input 	      nOE;		// output enable

   //---------------------------------------------------------------------------

   reg [DQ_size/2:1]  memL[0:mem_size-1],
                      memU[0:mem_size-1];

   wire 	      LB, UB, CE, WE;
   wire [DQ_size:1]   DQi;
   wire [A_size-1:0]  A_t;
   
   
   // tRC  cycle 
   // tAA  adress access
   // tAC  chip enable access (nCE)
   // tOE  output enable to output valid (nOE)
   // tOH  output hold from address change
   // tLZ  chip enable to output in low Z (nCE)
   // tOLZ output enable to output in low Z (nOE)
   // tHZ  chip disable to output in high Z (nCE)
   // tOHZ output disable to output in high Z (nOE)

   // tWLZ output low Z from end of write
   // tWHZ write to output in high Z

   parameter 	      
`ifdef  SP8NS       // GS71216/GS72116/GS74116
     tRC  = 8,   tAA  = 8,   tAC  = 8,   tAB = 3.5,  tOE  = 3.5, 
          tOH  = 3,   tLZ  = 3,   tOLZ = 0,   tBLZ = 0,   tHZ  = 4,   
	  tOHZ = 3.5, tBHZ = 3.5, tWLZ = 3,   tWHZ = 3.5, tWOH=1;
`endif
`ifdef  SP9NS       // GS71216
   tRC  = 9,   tAA  = 9,   tAC  = 9,   tAB = 3.5,  tOE  = 3.5, 
          tOH  = 3,   tLZ  = 3,   tOLZ = 0,   tBLZ = 0,   tHZ  = 4,   
          tOHZ = 3.5, tBHZ = 3.5, tWLZ = 3,   tWHZ = 3.5, tWOH=1;
`endif
`ifdef  SP10NS      // GS71116/GS72116/GS74116
   tRC  = 10,  tAA  = 10,  tAC  = 10,  tAB = 4,    tOE  = 4, 
          tOH  = 3,   tLZ  = 3,   tOLZ = 0,   tBLZ = 0,   tHZ  = 5,   
          tOHZ = 4,   tBHZ = 4,   tWLZ = 3,   tWHZ = 4, tWOH=1;
`endif
`ifdef  SP12NS      // GS71116/GS72116/GS74116
   tRC  = 12,  tAA  = 12,  tAC  = 12,  tAB = 5,    tOE  = 5, 
          tOH  = 3,   tLZ  = 3,   tOLZ = 0,   tBLZ = 0,   tHZ  = 6,   
          tOHZ = 5,   tBHZ = 5,   tWLZ = 3,   tWHZ = 5, tWOH=1;
`endif
`ifdef  SP15NS      // GS71116/GS72116/GS74116
   tRC  = 15,  tAA  = 15,  tAC  = 15,  tAB = 6,    tOE  = 6, 
          tOH  = 3,   tLZ  = 3,   tOLZ = 0,   tBLZ = 0,   tHZ  = 7,   
          tOHZ = 6,   tBHZ = 6,   tWLZ = 3,   tWHZ = 6, tWOH=1;
`endif


   // tWC  write cycle time
   // tAW  address valid to end of write
   // tCW  chip enable to end of write
   // tDW  data setup time
   // tDH  data hold time
   // tWP  write pulse width
   // tAS  address setup time
   // tWR  write recovery time (nWE)
   // tWR1 write recovery time (nCE)

`ifdef  SP8NS       // GS71216/GS72116/GS74116
   specify
      specparam
        tWC  = 8,   tAW  = 5.5, tCW  = 5.5, tBW = 5.5,  tDW  = 4,   
        tDH  = 0,   tWP  = 5.5, tAS  = 0,   tWR  = 0,   tWR1 = 0;
`endif
`ifdef  SP9NS       // GS71216
      specify
         specparam
           tWC  = 9,   tAW  = 5.5, tCW  = 5.5, tBW = 5.5,  tDW  = 4,   
           tDH  = 0,   tWP  = 5.5, tAS  = 0,   tWR  = 0,   tWR1 = 0;
`endif
`ifdef  SP10NS      // GS71116/GS72116/GS74116
	 specify
            specparam
              tWC  = 10,  tAW  = 7,   tCW  = 7,   tBW = 7,    tDW  = 5,   
              tDH  = 0,   tWP  = 7,   tAS  = 0,   tWR  = 0,   tWR1 = 0;
`endif
`ifdef  SP12NS      // GS71116/GS72116/GS74116
	    specify
               specparam
		 tWC  = 12,  tAW  = 8,   tCW  = 8,   tBW = 8,    tDW  = 6,   
		 tDH  = 0,   tWP  = 8,   tAS  = 0,   tWR  = 0,   tWR1 = 0;
`endif
`ifdef  SP15NS      // GS71116/GS72116/GS74116
	       specify
		  specparam
		    tWC  = 15,  tAW  = 10,  tCW  = 10,  tBW = 10,    tDW  = 7,   
		    tDH  = 0,   tWP  = 10,  tAS  = 0,   tWR  = 0,   tWR1 = 0;
`endif

		  // reading
		  $width ( posedge A, tRC );          // read address cycle
		  $width ( negedge A, tRC );
		  // writing
		  $width ( posedge A, tWC );          // write address cycle
		  $width ( negedge A, tWC );
		  
		  $width ( negedge nWE, tWP );        // we width         $setuphold  ( negedge nWE, A, tAS, tWP );       //    address
		  $setup ( negedge nWE, A, tAS);      //    address
		  $hold  ( posedge nWE, A, tWR);      //    address
		  $setup ( posedge nWE, A, tAW);
		  $hold  ( posedge nWE, A, tWR);
		  $setup ( DQ, posedge nWE,tDW);      //    data
		  $hold  ( DQ, posedge nWE,tDH);      //    data
		  
		  $width ( negedge nCE, tCW );        // ce width
		  $setup ( negedge nCE, A, tAS);      //    address
		  $hold  ( posedge nCE, A, tWR1);     //    address
		  $setup ( posedge nCE, A, tCW);
		  $hold  ( posedge nCE, A, tWR1);
		  $setup ( DQ, posedge nCE,tDW);      //    data
		  $hold  ( DQ, posedge nCE,tDH);      //    data
		  
	       endspecify

	       reg    triz_we,
		      triz_oe,
		      triz_lb,
		      triz_ub, temp_wr,
		      temp_lb, temp_ub, temp_oe, temp_we;
	       reg [DQ_size/2:1] memU_i_ce,
				 memL_i_ce,
				 memU_i_A,
				 memL_i_A,
				 memU_i,
				 memL_i;
	       wire [DQ_size/2:1] datau_o,
				  datal_o;
	       reg 		  nLB=0, nUB=0;
	       reg 		  acc_ce,
				  acc_oe,
				  acc_lb,
				  acc_ub,
				  acc_A, acc_A_load;
	       reg 		  data_mv_lb_A, data_mv_ub_A, data_mv_lb_oe, data_mv_ub_oe,
				  data_mv_lb, data_mv_ub, data_mv_A, data_mv_oe, data_mv_ce,
				  data_mv_oe_Z, data_mv_ub_oe_Z, data_mv_lb_oe_Z,
				  data_mv_lb_ce, data_mv_ub_ce;

	       always @(posedge nWE)
		 begin
		    temp_we <= 0;
		    triz_we <= 0;
		    temp_wr <= 0;
		    if (CE==0)
		      begin
			 if (LB==0) memL[A_t] <= DQi[8:1];
			 if (UB==0) memU[A_t] <= DQi[16:9];
		      end // if (!nCE)
		    if ((nOE==0) && (nCE==0) && ((nLB==0) || (nUB==0)) && (acc_A==0))
		      begin
			 acc_oe  <= 1;
			 acc_A_load <=1;
			 temp_oe <= 0;
			 triz_oe <= #tOLZ 0;
		      end // if (nWE & !nCE)
		    else
		      begin
			 acc_oe <= 0;
			 temp_oe <= 1;
			 triz_oe <= #tOLZ 1;
		      end // else: !if((nOE==0) && (nCE==0) && (nLB==0) && (nUB==0) && (acc_A==0))
		 end // always @ (posedge nWE)

   always @(posedge nCE)
     begin
	if (WE==0)
	  begin
	     if (LB==0) memL[A_t] <= DQi[8:1];
	     if (UB==0) memU[A_t] <= DQi[16:9];
	  end // if (!nWE)
	acc_oe <= 0;
	temp_oe <= #tOHZ 1;
	triz_oe <= #tOHZ 1;
     end // always @ (posedge nCE)

   always @ (A or negedge nOE or negedge nCE or negedge nLB or negedge nUB)
     begin
	if ((nWE==1) && (nCE==0) && (nOE==0) && ((nLB==0) || (nUB==0)))
	  begin
	     acc_A <= 1;
	     acc_A_load <= 1;
	  end // if ((nWE==1) && (nCE==0) && (nOE==0) && ((nLB==0) || (nUB==0)))
	else
	  begin
	     acc_A <= 0;
	     acc_A_load <= 1;
	  end // else: !if((nWE==1) && (nCE==0) && (nOE==0) && ((nLB==0) || (nUB==0)))
     end // always @ (A)

   always @ (negedge nOE or negedge nCE or negedge nLB or negedge nUB or posedge nWE)
     begin
	if ((nWE==1) && (nOE==0) && (nCE==0) && ((nLB==0) || (nUB==0)) && (acc_A==0))
	  begin
	     acc_oe  <= 1;
	     temp_oe <= #tOE 0;
	     triz_oe <= #tOE 0;
	  end // if (nWE & !nCE)
	else
	  acc_oe <= 0;
     end // always @ (negedge nOE)
   
   always @ (negedge nOE or negedge nCE or negedge nLB or negedge nUB)
     begin
	if ((nWE==1) && (nCE==0) && (nOE==0) && ((nLB==0) || (nUB==0)) && (acc_A==0))
	  acc_ce <= 1;
	else
	  acc_ce <= 0;
     end // always @ (negedge nCE)

   always @ (negedge nCE or negedge nOE or negedge nLB or negedge nUB)
     begin
	triz_lb <= #tBLZ 0;
	temp_lb <= 0;
	if ((nWE==1) && (nOE==0) && (nCE==0))
	  acc_lb <= 1;
	else
	  acc_lb <= 0;
     end // always @ (negedge nCE)

   always @ (negedge nOE or negedge nCE or negedge nLB or negedge nUB)
     begin
	triz_ub <= #tBLZ 0;
	temp_ub <= 0;
	if ((nWE==1) && (nOE==0) && (nCE==0))
	  acc_ub <= 1;
	else
	  acc_ub <= 0;
     end // always @ (negedge nCE)

   always @ (negedge nWE)
     begin
	temp_wr <= #tWOH 1;
	temp_we <= 1;
	triz_we <= 1;
	data_mv_lb <= 1;
	data_mv_ub <= 1;
	data_mv_lb <= #1 0;
	data_mv_ub <= #1 0;	 
     end // always @ (negedge nWE)

   always @ (posedge nOE)
     begin
	temp_oe <= 1;
	acc_oe  <= 0;
	triz_oe <= #tOHZ 1;
     end // always @ (posedge nOE)

   always @ (posedge nLB)
     begin
	temp_lb <= 1;
	triz_lb <= #tBHZ 1;
	acc_lb <= 0;
     end // always @ (posedge nLB)

   always @ (posedge nUB)
     begin
	temp_ub <= 1;
	triz_ub <= #tBHZ 1;
	acc_ub <= 0;
     end // always @ (posedge nUB)

   always @ (posedge temp_wr)
     begin
	data_mv_lb <= 1;
	data_mv_ub <= 1;
	data_mv_lb <= #1 0;
	data_mv_ub <= #1 0;	 
     end // always @ (posedge triz_we)

   always @ (posedge triz_we)
     begin
	if (temp_we==0 && temp_oe==0) 
	  triz_we <= 0;
     end // always @ (posedge triz_we)

   always @ (posedge triz_oe)
     begin
	if (temp_oe==0 && temp_we==0) 
	  begin
	     triz_oe <= 0;
	  end // if (temp_oe==0 || temp_we==0)
     end // always @ (posedge triz_oe)

   always @ (posedge triz_ub)
     begin
	if (temp_ub==0) 
	  triz_ub <= temp_ub;
     end // always @ (posedge triz_ub)

   always @ (posedge triz_lb)
     begin
	if (temp_lb==0) 
	  triz_lb <= temp_lb;
     end // always @ (posedge triz_lb)

   always @ (posedge acc_A or posedge acc_A_load or posedge acc_oe or posedge acc_ce)
     begin
	if (acc_A)
	  begin
	     if (!nLB)
	       begin
		  data_mv_lb_A <= #(tOH+1) 0;
		  memL_i_A <= #tOH 'hx;
		  memL_i_A <= #tAA memL[A];
		  data_mv_lb_A <= #tAA 1;
	       end // if (acc_lb==1)
	     if (!nUB)
	       begin
		  data_mv_ub_A <= #(tOH+1) 0;
		  memU_i_A <= #tOH 'hx;
		  memU_i_A <= #tAA memU[A];
		  data_mv_ub_A <= #tAA 1;
	       end // if (acc_ub==1)
	     acc_A  <= #2 0;
	     data_mv_A <= #tOH 0;
	     data_mv_A <= #tAA 1;
	  end // if (acc_A==1)
	if (acc_A_load)
	  begin
	     if (!nLB)
	       begin
		  data_mv_lb_A <= #(tOH+1) 0;
		  memL_i_A <= #tOH 'hx;
		  memL_i_A <= #tAA memL[A];
		  data_mv_lb_A <= #tAA 1;
	       end // if (acc_lb==1)
	     if (!nUB)
	       begin
		  data_mv_ub_A <= #(tOH+1) 0;
		  memU_i_A <= #tOH 'hx;
		  memU_i_A <= #tAA memU[A];
		  data_mv_ub_A <= #tAA 1;
	       end // if (acc_ub==1)
	     acc_A_load  <= #2 0;
	     data_mv_A <= #tOH 0;
	     data_mv_A <= #tAA 1;
	  end // if (acc_A==1)
     end // always @ (posedge acc_A or posedge acc_oe or posedge acc_ce)

   always @ (posedge acc_A or posedge acc_oe or posedge acc_ce)
     begin
	if (acc_oe==1)
	  begin
	     if (!nLB)
	       begin
		  data_mv_lb_oe <= #tOE 1;
		  data_mv_lb_oe_Z <= #tOLZ 1;
	       end // if (!nLB)
	     if (!nUB)
	       begin
		  data_mv_ub_oe <= #tOE 1;
		  data_mv_ub_oe_Z <= #tOLZ 1;
	       end // if (!nUB)
	     acc_oe <= #2 0;
	     data_mv_oe <= #tOE 1;
	     data_mv_oe_Z <= #tOLZ 1;
	  end // if (acc_oe==1)	    
	if (acc_oe==1 && acc_A==0)
	  begin
	     if (!nLB)
	       begin
		  data_mv_lb_oe <= #tOE 1;
		  data_mv_lb_oe_Z <= #tOLZ 1;
	       end // if (!nLB)
	     if (!nUB)
	       begin
		  data_mv_ub_oe <= #tOE 1;
		  data_mv_ub_oe_Z <= #tOLZ 1;
	       end // if (!nUB)
	     acc_oe <= #2 0;
	     data_mv_oe <= #tOE 1;
	     data_mv_oe_Z <= #tOLZ 1;
	  end // if (acc_oe==1)	    
     end // always @ (posedge acc_A or posedge acc_oe or posedge acc_ce)

   always @ (posedge acc_A or posedge acc_oe or posedge acc_ce)
     begin
	if (acc_ce==1)
	  begin
	     if (!nLB)
	       begin
		  memL_i_ce <= #tLZ 'hx;
		  memL_i_ce <= #tAC memL[A];
		  data_mv_lb <= #tAC 1;
	       end // if (acc_lb==1)
	     if (!nUB)
	       begin
		  memU_i_ce <= #tLZ 'hx;
		  memU_i_ce <= #tAC memU[A];
		  data_mv_ub <= #tAC 1;
	       end // if (acc_ub==1)
	     acc_ce <= #2 0;
	  end // if (acc_ce==1)
     end // always @ (posedge acc_A or posedge acc_oe or posedge acc_ce)

   always @ (data_mv_A)
     begin
	if (data_mv_lb_A)
	  begin
	     data_mv_lb <= 1;
	     memL_i <= memL_i_A;
	  end // if (data_mv_lb_oe)
	if (data_mv_ub_A)
	  begin
	     data_mv_ub <= 1;
	     memU_i <= memU_i_A;
	  end // if (data_mv_ub_oe)
     end // always @ (posedge data_mv_A or posedge data_mv_oe...
   always @ (posedge data_mv_oe)
     begin
	if (data_mv_lb_oe)
	  begin
	     data_mv_lb <= 1;
	     memL_i <= memL_i_A;
	  end // if (data_mv_lb_oe)
	if (data_mv_ub_oe)
	  begin
	     data_mv_ub <= 1;
	     memU_i <= memU_i_A;
	  end // if (data_mv_ub_oe)
	data_mv_oe <= #1 0;
	data_mv_lb_oe <= #1 0;
	data_mv_ub_oe <= #1 0;
     end // always @ (posedge data_mv_A or posedge data_mv_oe...
   always @ ( posedge data_mv_ce)
     begin
	if (data_mv_lb_oe_Z)
	  begin
	     data_mv_lb <= 1;
	     memL_i <= #tOLZ 'hz;
	  end // if (data_mv_lb_oe_Z)
	if (data_mv_ub_oe_Z)
	  begin
	     data_mv_ub <= 1;
	     memU_i <= #tOLZ 'hz;
	  end // if (data_mv_ub_oe_Z)
	data_mv_oe_Z <= #1 0;
	data_mv_lb_oe_Z <= #1 0;
	data_mv_ub_oe_Z <= #1 0;
     end // always @ (posedge data_mv_A or posedge data_mv_oe...
   always @ ( posedge data_mv_oe_Z)
     begin
	if (data_mv_lb_ce)
	  begin
	     data_mv_lb <= 1;
	     memL_i <= memL_i_ce;
	  end // if (data_mv_lb_oe)
	if (data_mv_ub_ce)
	  begin
	     data_mv_ub <= 1;
	     memU_i <= memU_i_ce;
	  end // if (data_mv_ub_oe)
	data_mv_ce <= #1 0;
	data_mv_lb_ce <= #1 0;
	data_mv_ub_ce <= #1 0;
     end // always @ (posedge data_mv_A or posedge data_mv_oe...
   
   assign  datau_o = (!data_mv_ub)? 'hz: memU_i;
   assign  datal_o = (!data_mv_lb)? 'hz: memL_i;

   assign  #1 LB = nLB;
   assign  #1 UB = nUB;
   assign  #1 CE = nCE;
   assign  #1 WE = nWE;
   assign  #1 DQi = DQ;
   assign  #1 A_t = A;
   
   assign  DQ[16:9] = (triz_we|triz_oe|triz_ub)? 'hz: datau_o;
   assign  DQ[8:1]  = (triz_we|triz_oe|triz_lb)? 'hz: datal_o;
   
endmodule

